//////////////////////////////////////////////////////////////////////////////
//                                                                          //
//  Copyright (C) 2006 David Lubicz <david.lubicz@math.univ-rennes1.fr>     //
//  Copyright (C) 2006 David Kohel <kohel@maths.usyd.edu.au>                //
//  Copyright (C) 2006 Robert Carls <carls@maths.usyd.edu.au>               //
//                                                                          //
//  Distributed under the terms of the GNU General Public License (GPL)     //
//                  http://www.gnu.org/licenses/                            //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

/*
We implement an algorithm for canonical lifting of ordinary
   theta null points
*/

/////////////////////////////////////////////////////////////////////////
declare verbose CanonicalThetaNull, 2;
/////////////////////////////////////////////////////////////////////////
   

/////////////////////////////////////////////////////////////////////////
// Manipulating theta null points
/////////////////////////////////////////////////////////////////////////

function NormaliseProjectivePointFirstVariable(a)
    c := Eltseq(a)[1];
    assert IsInvertible(c);
    return Vector([ entry * c^-1 : entry in Eltseq(a) ]); 
end function;

//////////////////////////////////////////////////////////////////////////////
// Implementation of the Thomae formulas
//////////////////////////////////////////////////////////////////////////////

intrinsic ThomaeConstants(H::CrvHyp) -> ModTupRngElt
    {The 2-torsion part of the theta null point of the Jacobian of the curve H
    is computed. Note that this requires the choice of a 4-theta structure.}

    R := BaseRing(H);

    assert IsField(R);
    assert Characteristic(R) ne 2;

    assert Genus(H) eq 2;

    C := SimplifiedModel(H);
    P1 := PolynomialRing(R);
    f := HyperellipticPolynomials(C);    

    assert Degree(f) eq 5;

    RSplit := SplittingField(f);
    P2<s>, i := ChangeRing(P1,RSplit);
    e1, e2, e3, e4, e5 := Explode([ ei[1] : ei in Roots(i(f)) ]);

    g := (s^4-(e1-e2))*(s^4-(e2-e3))*(s^4-(e3-e4))*(s^4-(e4-e5))
         *(s^4-(e1-e4))*(s^4-(e2-e5))*(s^4-(e1-e3))*(s^4-(e1-e5))
         *(s^4-(e3-e5))*(s^4-(e2-e4));
    RSplit1 := SplittingField(g);

    q1 := ((e2-e5)/(e2-e4))*((e3-e4)/(e3-e5));
    q2 := ((e1-e4)/(e1-e5));
    q3 := ((e1-e2)/(e1-e3));

    f1 := Sqrt(Sqrt(RSplit1!(q1)));
    f2 := Sqrt(Sqrt(RSplit1!(q2)));
    f3 := Sqrt(Sqrt(RSplit1!(q3)));
    
    return Vector([ 1, f1*f2, f2*f3, f1*f3 ]);
end intrinsic;

intrinsic HyperellipticCurveFromThomaeConstants(T::ModTupRngElt) -> CrvHyp
    {A hyperelliptic curve is constructed from given Thomae constants.}

    R := BaseRing(Parent(T));    
    a00, a02, a20, a22 := Explode(Eltseq(NormaliseProjectivePointFirstVariable(T)));

    e1 := R!0;
    e2 := R!1;
    e3 := (a02/(a22*a20))^2;
    PR<t> := PolynomialRing(R);
    f := (a02^2*a22^2-a20^2) / (a02^4*a20^2) * t^2 - (a02^4-a20^4+a22^4-1)
         / (a02^2*a20^2*a22^2) * t + (a02^2*a22^2-a20^2) / (a20^2*a22^4);
    RR := SplittingField(f);
    PRR<s> := PolynomialRing(RR);
    e4 := Roots((a02^2*a22^2-a20^2) / (a02^4*a20^2) * s^2 - (a02^4-a20^4+a22^4-1)
         / (a02^2*a20^2*a22^2) * s + (a02^2*a22^2-a20^2) / (a20^2*a22^4))[1][1];
    e5 := e4 * (a22 / (a20*a02))^2;
    return HyperellipticCurve((s-e1)*(s-e2)*(s-e3)*(s-e4)*(s-e5));

end intrinsic;

intrinsic HyperellipticPolynomialFromThomaeConstantsOverRing(T::ModTupRngElt) -> CrvHyp
    {A hyperelliptic polynomial is constructed from given Thomae constants over a ring.}

    R := BaseRing(Parent(T));
    a00, a02, a20, a22 := Explode(Eltseq(NormaliseProjectivePointFirstVariable(T)));
    e1 := R!0;
    e2 := R!1;
    e3 := (a02 div (a22*a20))^2;
    a := (a02^2*a22^2-a20^2) div (a02^4*a20^2);
    b := (a02^4-a20^4+a22^4-1) div (a02^2*a20^2*a22^2);
    c := (a02^2*a22^2-a20^2) div (a20^2*a22^4);
    d := Sqrt(b^2-4*a*c);
    e4 := (1 div (2*a))*(-b+d);
    e5 := e4 * (a22 div (a20*a02))^2;
    PR<s> := PolynomialRing(R);
    return (s-e1)*(s-e2)*(s-e3)*(s-e4)*(s-e5);

end intrinsic;

intrinsic TwoTorsionPartOfFourThetaNullPoint(v::ModTupRngElt) -> ModTupRngElt
    {}
    return NormaliseProjectivePointFirstVariable(Vector(Eltseq(v)[[1..4]]));
end intrinsic;

intrinsic CompleteFourThetaNullPointOverField(A::ModTupRngElt) -> ModTupRngElt
    {A theta null point is computed that has 2-torsion part equal to the given
    point. Note that this depends on some choices. Possibly one has to extend
    the base field.}

    R := BaseRing(Parent(A));      

    assert IsField(R);
    assert not Characteristic(R) eq 2;

    A00, A02, A20, A22 := Explode(Eltseq(A));

    P<u> := PolynomialRing(R);

    h := (u^2-(A00^2+A02^2+A20^2+A22^2)/2)
	*(u^2-(A00*A02+A20*A22))
	*(u^2-(A00*A20+A02*A22))
        *(u^2-(A00*A22+A02*A20));

    RR := SplittingField(h);

    G1 := Sqrt(RR!((A00^2+A02^2+A20^2+A22^2)/2));
    G2 := Sqrt(RR!(A00*A02+A20*A22));
    G3 := Sqrt(RR!(A00*A20+A02*A22));
    G4 := Sqrt(RR!(A00*A22+A02*A20));
    
    S1 := G1 * G2;
    S2 := G1 * G3;
    S3 := G1 * G4;
    T1 := G3 * G4;
    T2 := G2 * G4;
    T3 := G2 * G3;

    PP<w> := PolynomialRing(RR);

    g := (w^2-(S1+T1))
        *(w^2-(S1-T1))
        *(w^2-(S2+T2))
        *(w^2-(S2-T2))
        *(w^2-(S3+T3))
        *(w^2-(S3-T3));

    RRR := SplittingField(g);

    W1 := Sqrt(RRR!(S1+T1));
    W2 := Sqrt(RRR!(S1-T1));
    W3 := Sqrt(RRR!(S2+T2));
    W4 := Sqrt(RRR!(S2-T2));
    W5 := Sqrt(RRR!(S3+T3));
    W6 := Sqrt(RRR!(S3-T3));

    A01 := (W1+W2)/2;
    A21 := (W1-W2)/2;
    A10 := (W3+W4)/2;
    A12 := (W3-W4)/2;
    A11 := (W5+W6)/2;
    A13 := (W5-W6)/2;

    return  NormaliseProjectivePointFirstVariable(
	Vector([ A00, A02, A20, A22, A01, A21, A10, A12, A11, A13 ]));
end intrinsic;

function HenselSqrt(x,y0)
    // We could do something more intelligent here, but this 
    // works to extract the one bit of information.
    y := Sqrt(x);
    if Valuation(y-y0) gt 0 then
	return y;
    end if;
    assert Valuation(y+y0) gt 0;
    return -y;
end function;

function CompleteFourThetaNullPointHensel(A,a)
    /*
    Does the same as CompleteFourThetaNullPointOverField() 
    but over a ring.
    Here it is assumed that the full theta null point is defined over
    the same ring as the 2-torsion part.
    */ 

    R := BaseRing(Parent(A));

    assert IsInvertible(R!2);

    A00, A02, A20, A22 := Explode(Eltseq(A));
    a00, a02, a20, a22, a01, a21, a10, a12, a11, a13 :=
	Explode(ChangeUniverse(Eltseq(a),R));

    w1 := a01 + a21;
    w2 := a01 - a21;
    w3 := a10 + a12;
    w4 := a10 - a12;
    w5 := a11 + a13;
    w6 := a11 - a13; 

    s1 := (w1^2 + w2^2) div 2;
    t1 := (w1^2 - w2^2) div 2;
    s2 := (w3^2 + w4^2) div 2;
    t2 := (w3^2 - w4^2) div 2;
    s3 := (w5^2 + w6^2) div 2;
    t3 := (w5^2 - w6^2) div 2;

    g1 := Sqrt((a00^2+a02^2+a20^2+a22^2) div 2);
    g2 := s1 div g1; // assert Valuation(g2^2 - (a00*a02+a20*a22)) gt 0;
    g3 := s2 div g1; // assert Valuation(g3^2 - (a00*a20+a02*a22)) gt 0;
    g4 := s3 div g1; // assert Valuation(g4^2 - (a00*a22+a02*a20)) gt 0;
    // assert Valuation(t1-g3*g4) gt 0;
    // assert Valuation(t2-g2*g4) gt 0;
    // assert Valuation(t3-g2*g3) gt 0;
    
    G1 := HenselSqrt((A00^2+A02^2+A20^2+A22^2) div 2,g1);
    G2 := HenselSqrt(A00*A02+A20*A22,g2);
    G3 := HenselSqrt(A00*A20+A02*A22,g3);
    G4 := HenselSqrt(A00*A22+A02*A20,g4);
    
    S1 := G1 * G2;
    S2 := G1 * G3;
    S3 := G1 * G4;
    T1 := G3 * G4;
    T2 := G2 * G4;
    T3 := G2 * G3;

    W1 := HenselSqrt(S1+T1,w1);
    W2 := HenselSqrt(S1-T1,w2);
    W3 := HenselSqrt(S2+T2,w3);
    W4 := HenselSqrt(S2-T2,w4);
    W5 := HenselSqrt(S3+T3,w5);
    W6 := HenselSqrt(S3-T3,w6);

    A01 := (W1+W2) div 2;
    A21 := (W1-W2) div 2;
    A10 := (W3+W4) div 2;
    A12 := (W3-W4) div 2;
    A11 := (W5+W6) div 2;
    A13 := (W5-W6) div 2;

    return  NormaliseProjectivePointFirstVariable(
	Vector([ A00, A02, A20, A22, A01, A21, A10, A12, A11, A13 ]));

end function;

function CompleteFourThetaNullPointOverRing(A)
    /*
    Does the same as CompleteFourThetaNullPointOverField() but
    over a ring.
    Here it is assumed that the full theta null point is defined over the same
    ring as the 2-torsion part.
    */ 

    R := BaseRing(Parent(A));

    assert IsInvertible(R!2);

    A00, A02, A20, A22 := Explode(Eltseq(A));

    G1 := Sqrt((A00^2+A02^2+A20^2+A22^2) div 2);
    G2 := Sqrt(A00*A02+A20*A22);
    G3 := Sqrt(A00*A20+A02*A22);
    G4 := Sqrt(A00*A22+A02*A20);
    
    S1 := G1 * G2;
    S2 := G1 * G3;
    S3 := G1 * G4;
    T1 := G3 * G4;
    T2 := G2 * G4;
    T3 := G2 * G3;

    W1 := Sqrt(S1+T1);
    W2 := Sqrt(S1-T1);
    W3 := Sqrt(S2+T2);
    W4 := Sqrt(S2-T2);
    W5 := Sqrt(S3+T3);
    W6 := Sqrt(S3-T3);

    A01 := (W1+W2) div 2;
    A21 := (W1-W2) div 2;
    A10 := (W3+W4) div 2;
    A12 := (W3-W4) div 2;
    A11 := (W5+W6) div 2;
    A13 := (W5-W6) div 2;

    return  NormaliseProjectivePointFirstVariable(
	Vector([ A00, A02, A20, A22, A01, A21, A10, A12, A11, A13 ]));

end function;

intrinsic ThetaNullPoint(H::CrvHyp) -> ModTupRngElt
    {The theta null point of the Jacobian of the curve H is computed.}
    A := ThomaeConstants(H);
    return CompleteFourThetaNullPointOverField(A);
end intrinsic;

///////////////////////////////////////////////////////////////////////////////
// Modular Equations
///////////////////////////////////////////////////////////////////////////////

function GetRiemannRelations(P)

     return [ 2*(P.5^2+P.6^2)^2-(P.1^2+P.2^2+P.3^2+P.4^2)*(P.1*P.2+P.3*P.4),
	      2*(P.7^2+P.8^2)^2-(P.1^2+P.2^2+P.3^2+P.4^2)*(P.1*P.3+P.2*P.4),
	      2*(P.9^2+P.10^2)^2-(P.1^2+P.2^2+P.3^2+P.4^2)*(P.1*P.4+P.3*P.2),
	      4*P.5^2*P.6^2-(P.1*P.3+P.2*P.4)*(P.1*P.4+P.2*P.3),
	      4*P.7^2*P.8^2-(P.1*P.2+P.3*P.4)*(P.1*P.4+P.2*P.3),
	      4*P.9^2*P.10^2-(P.1*P.2+P.3*P.4)*(P.1*P.3+P.2*P.4),
	      (P.1^2+P.2^2+P.3^2+P.4^2)*P.10*P.9-(P.8^2+P.7^2)*(P.5^2+P.6^2),
	      (P.1^2+P.2^2+P.3^2+P.4^2)*P.5*P.6-(P.8^2+P.7^2)*(P.9^2+P.10^2),
	      (P.1^2+P.2^2+P.3^2+P.4^2)*P.7*P.8-(P.5^2+P.6^2)*(P.9^2+P.10^2),
	      (P.2*P.3+P.1*P.4)*P.9*P.10-2*P.5*P.7*P.6*P.8,
	      (P.2*P.3+P.1*P.4)*(P.5^2+P.6^2)-2*P.7*P.8*(P.9^2+P.10^2),
	      (P.2*P.3+P.1*P.4)*(P.7^2+P.8^2)-2*P.5*P.6*(P.9^2+P.10^2),
	      (P.1*P.3+P.4*P.2)*P.7*P.8-2*P.9*P.10*P.6*P.5,
	      (P.1*P.3+P.4*P.2)*(P.10^2+P.9^2)-2*(P.7^2+P.8^2)*P.6*P.5,
	      (P.1*P.3+P.4*P.2)*(P.6^2+P.5^2)-2*(P.7^2+P.8^2)*P.9*P.10,
	      (P.1*P.2+P.3*P.4)*P.6*P.5-2*P.9*P.10*P.7*P.8,
	      (P.1*P.2+P.3*P.4)*(P.8^2+P.7^2)-2*P.9*P.10*(P.5^2+P.6^2),
	      (P.1*P.2+P.3*P.4)*(P.9^2+P.10^2)-2*P.7*P.8*(P.5^2+P.6^2),
	      P.5*P.6*(P.5^2+P.6^2)-P.7*P.8*(P.7^2+P.8^2),
	      P.5*P.6*(P.5^2+P.6^2)-P.9*P.10*(P.9^2+P.10^2) ];

end function;

function GetCorrespondenceRelations(prime,PP)
     if (prime eq 2) then
         return [
		 PP.1^2*(PP.15*PP.11+PP.15*PP.12+PP.16*PP.13+PP.16*PP.14)
		   -PP.5^2*(PP.11^2+PP.12^2+PP.13^2+PP.14^2),
		 PP.1^2*(PP.17*PP.11+PP.18*PP.12+PP.17*PP.13+PP.18*PP.14)
                   -PP.7^2*(PP.11^2+PP.12^2+PP.13^2+PP.14^2),
		 PP.1^2*(PP.19*PP.11+PP.20*PP.12+PP.20*PP.13+PP.19*PP.14)
                   -PP.9^2*(PP.11^2+PP.12^2+PP.13^2+PP.14^2) ];
     else
         return [
	     PP.11*PP.2+PP.12*PP.1+PP.13*PP.4+PP.3*PP.14-2*(PP.5*PP.15+PP.16*PP.6),
	     PP.11*PP.3+PP.1*PP.13+PP.12*PP.4+PP.14*PP.2-2*(PP.17*PP.7+PP.18*PP.8),
	     PP.11*PP.4+PP.14*PP.1+PP.12*PP.3+PP.13*PP.2-2*(PP.10*PP.20+PP.19*PP.9) ];
     end if;
end function;


///////////////////////////////////////////////////////////////////////
// Some helpful functions
///////////////////////////////////////////////////////////////////////

function MinValuationVector(a)
    return Min([ Valuation(entry) : entry in Eltseq(a) ]);
end function;

function DivVector(a,n)
     return Vector([ entry div n : entry in Eltseq(a) ]);
end function;

function CoerceVector(a,S)
     return Vector([ S!entry : entry in Eltseq(a) ]);
end function;

function CoerceMatrix(M,S)
    return Matrix([[ S!M[i,j] : j in [1..Ncols(M)]] : i in [1..Nrows(M)]]);
end function;

function FrobVector(A, k)
     R := BaseRing(Parent(A));
     case Category(R):
     when FldFin:
         p := Characteristic(R);
         return Vector([ entry^(p^k) : entry in Eltseq(A) ]);
     when RngPadResExt:
         return Vector([ GaloisImage(c,k) : c in Eltseq(A) ]);
     end case;
     assert false;
end function;

function FrobMatrix(A, k)
     R := BaseRing(Parent(A));
     case Category(R):
     when FldFin:
	 p := Characteristic(R);
         return Parent(A)![ c^(p^k) : c in Eltseq(A) ];
     when RngPadResExt:
         return Parent(A)![ GaloisImage(c,k) : c in Eltseq(A) ];
     end case;
     assert false;
end function;

function EvaluateMatrix(M,x)
    c := Eltseq(x);
    return Matrix([ [
	Evaluate(M[i,j],c) : j in [1..Ncols(M)] ] : i in [1..Nrows(M)] ]);
end function;

function CatVectors(A,B)
    return Vector(Eltseq(A) cat Eltseq(B));
end function;


/////////////////////////////////////////////////////////////////////////
// Checking for valid theta null point
/////////////////////////////////////////////////////////////////////////

intrinsic IsValidFourThetaNullPoint(a::ModTupRngElt) -> BoolElt
    {It is checked if the given theta null point satisfies the Riemann relations.}
    R := BaseRing(Parent(a));
    P := PolynomialRing(R,10);    
    Rels := GetRiemannRelations(P);
    return &and[ Evaluate(f,Eltseq(a)) eq 0 : f in Rels ];
end intrinsic;

intrinsic SatisfiesCorrespondenceRelations(a::ModTupRngElt,prime::RngIntElt) -> BoolElt
    {It is checked if the given theta null point satisfies the correspondence relations.}
    R := BaseRing(Parent(a));
    P := PolynomialRing(R,20);
    Rels := GetCorrespondenceRelations(prime,P);
    b := FrobVector(a,1);
    c := CatVectors(a,b);
    return &and[ Evaluate(f,Eltseq(c)) eq 0 : f in Rels ];
end intrinsic;


////////////////////////////////////////////////////////////////////////
// The definition of Phi and derivatives
//////////////////////////////////////////////////////////////////////

function Phi(ModEQs,x,y)
    a := Eltseq(x);
    b := Eltseq(y);
    c := a cat b; 
    return Vector([ Evaluate(f,c) : f in ModEQs ]);
end function;

function DPhi0(DModEQs,x,y)
    return EvaluateMatrix(DModEQs,CatVectors(x,y));
end function;

function DerivativeLocalInverse(v,DRiem)
    R := BaseRing(Parent(v));

    D := EvaluateMatrix(DRiem,v);

    DPsi1 := ColumnSubmatrix(D,1,3);
    DPsi2 := ColumnSubmatrix(D,4,6);

    E, T := EchelonForm(DPsi2); assert RowSubmatrix(E,1,6) eq 1;

    U := IdentityMatrix(R,3);
    Dg := -RowSubmatrix(T,6)*DPsi1;
    return VerticalJoin(U,Dg);
end function;


/////////////////////////////////////////////////////////////////////////////////
// The lifting algorithm
/////////////////////////////////////////////////////////////////////////////////

intrinsic SolveArtinSchreierEquation(M::Mtrx,V::ModTupRngElt)
    -> ModTupRngElt
    {This function solves an equation of the form sigma(x)+Mx+V=0 for a vector x.
     The matrix M has to be equal zero over the residue field.}

    SS0 := BaseRing(Parent(M));

    KK0 := ResidueClassField(SS0);
    char := Characteristic(KK0);
    d := Degree(KK0);

    assert SS0 eq BaseRing(Parent(V));

    N := NumberOfRows(M);

    assert N eq NumberOfColumns(M);
    assert N eq Rank(Parent(V));

    prec := Precision(SS0);

    if prec eq 1 then
        return FrobVector(-V,d-1);
    end if;

    r := (prec+1) div 2;

    SS1 := ChangePrecision(SS0,r);
    M1 := Matrix([[ SS1!(M[i,j]) : j in [1..N] ] : i in [1..N]]);
    V1 := Vector([ SS1!entry : entry in Eltseq(V) ]);

    x := CoerceVector(SolveArtinSchreierEquation(M1,V1),SS0);

    V2 := CoerceVector(DivVector(FrobVector(x,1)+x*Transpose(M)+V,char^r),SS1);
    delta := CoerceVector(SolveArtinSchreierEquation(M1,V2),SS0);

    return x + char^r * delta;

end intrinsic;

function NewtonThetaLift(Xl4,ModEQs,DModEQs,Riem,DRiem)

    SS := BaseRing(Parent(Xl4));
    prec := Precision(SS);
    if prec eq 1
        then return Xl4;
    end if;
    KK := ResidueClassField(SS);
    char := Characteristic(KK);

    Xl4_prev := Xl4;
    Yl4_prev := FrobVector(Xl4_prev,1);
    Xl2_prev := TwoTorsionPartOfFourThetaNullPoint(Xl4);

    r := (prec+1) div 2;

//    printf "REDUCING TO LOWER PRECISION: %o\n", r;
    
    S1 := ChangePrecision(SS,r); AssignNames(~S1,["t"]);
    Xl4_half_S1 := CoerceVector(Xl4_prev,S1);
    Xl4_half_S1 := NewtonThetaLift(Xl4_half_S1,ModEQs,DModEQs,Riem,DRiem);
    Yl4_half_S1 := FrobVector(Xl4_half_S1,1);
//    printf "LIFTED HALFWAY TO PRECISION: %o\n", r;

    Xl4_half := CompleteFourThetaNullPointHensel(
	CoerceVector(Xl4_half_S1,SS),Xl4_prev);
    Yl4_half := FrobVector(Xl4_half,1);

    DX := DerivativeLocalInverse(Xl4_half_S1,DRiem);
    DY := FrobMatrix(DX,1);
    /*
    DER := EvaluateMatrix(
	DModEQs,CatVectors(Xl4_half,Yl4_half)) * DiagonalJoin(DX,DY);
    A := ColumnSubmatrix(DER,1,3); 
    B := ColumnSubmatrix(DER,4,3); 
    */
    tyme := Cputime();
    MatS := MatrixAlgebra(S1,3);
    DModEQsX := ColumnSubmatrix(DModEQs,1,9);
    A := MatS!(EvaluateMatrix(DModEQsX,CatVectors(Xl4_half_S1,Yl4_half_S1)) * DX);
    DModEQsY := ColumnSubmatrix(DModEQs,10,9);
    B := MatS!(EvaluateMatrix(DModEQsY,CatVectors(Xl4_half_S1,Yl4_half_S1)) * DY);
    vprint CanonicalThetaNull : "A,B-Matrix computations:", Cputime(tyme);
    
//    print "WORKING PRECISION:", Precision(S1);
    print "Precision:", Precision(SS);

    V := CoerceVector(DivVector(Phi(ModEQs,Xl4_half,Yl4_half),char^r),S1);

    assert CoerceMatrix(A,KK) eq ZeroMatrix(KK,3,3);

    bool, M := IsInvertible(B); assert bool;
    delta := CoerceVector(SolveArtinSchreierEquation(M*A,V*Transpose(M)),SS);

    if GetVerbose("CanonicalThetaNull") gt 0 then
	tyme := Cputime();
	X := EvaluateMatrix(DRiem,Xl4_half_S1);
	assert &and[ X[i]*DX eq 0 : i in [1..Nrows(DX)] ];
	long := true;
	if long then
	    delta_x := CoerceVector(delta * Transpose(DX),S1);
	    delta_y := FrobVector(delta_x,1);
	    DModX := EvaluateMatrix(DModEQsX,CatVectors(Xl4_half_S1,Yl4_half_S1));
	    DModY := EvaluateMatrix(DModEQsY,CatVectors(Xl4_half_S1,Yl4_half_S1));
	    assert CoerceVector(
		delta_y*Transpose(DModY) + delta_x*Transpose(DModX) + V,S1) eq 0;
	else
	    d_x := CoerceVector(delta,S1);
	    d_y := FrobVector(d_x,1);
	    B1 := CoerceMatrix(B,S1);
	    A1 := CoerceMatrix(A,S1);
	    V1 := CoerceVector(V,S1); 
	    assert d_y*Transpose(B1) + d_x*Transpose(A1) + V1 eq 0;
	end if;
	vprint CanonicalThetaNull : "Checking time #1:", Cputime(tyme);
    end if;
    
    tyme := Cputime();
    Xl2_half := TwoTorsionPartOfFourThetaNullPoint(Xl4_half);
    Xl2_full := Xl2_half + char^r * CatVectors(Vector([ SS!0 ]),delta);
    Xl4_full := CompleteFourThetaNullPointHensel(Xl2_full,Xl4_half);
    assert Xl2_full eq TwoTorsionPartOfFourThetaNullPoint(Xl4_full);
    vprint CanonicalThetaNull : "Full lifted computation:", Cputime(tyme);
    // Alternatively we could compute Xl4_full as follows:
    tyme := Cputime();
    full_delta := CatVectors(Vector([ S1!0 ]),delta * Transpose(DX));
    assert Xl4_full eq Xl4_half + char^r * CoerceVector(full_delta,SS);
    vprint CanonicalThetaNull : "Alternative computation:", Cputime(tyme);

    if GetVerbose("CanonicalThetaNull") gt 0 then
	tyme := Cputime();
	Yl4_prev := FrobVector(Xl4_prev,1);
	Yl4_full := FrobVector(Xl4_full,1);
	v_full := Vector([ Evaluate(f,Eltseq(Xl4_full)) : f in Riem ]);
	u_full := Phi(ModEQs,Xl4_full,Yl4_full);
	v_prev := Vector([ Evaluate(f,Eltseq(Xl4_prev)) : f in Riem ]);
	u_prev := Phi(ModEQs,Xl4_prev,Yl4_prev);
	q := char^r;
	print "Full u:", m where m := Min([ Valuation(c) : c in Eltseq(u_full) ]);
	print "Full v:", m where m := Min([ Valuation(c) : c in Eltseq(v_full) ]);
	print "Prev u:", m where m := Min([ Valuation(c) : c in Eltseq(u_prev) ]);
	print "Prev v:", m where m := Min([ Valuation(c) : c in Eltseq(v_prev) ]);
	assert &and[ c eq 0 : c in Eltseq(u_full) ];
	vprint CanonicalThetaNull : "Checking time #2:", Cputime(tyme);
    end if;
    return Xl4_full;
end function;    

intrinsic CanonicalLiftTwoThetaNullPoint(xl2::ModTupRngElt, prec:RngIntElt)
    -> ModTupRngElt
    {The 2-torsion part of the theta null point of the canonical lift
     is computed to precision prec.}

    assert prec ge 1;

    FF2 := BaseRing(Parent(xl2));
    assert Category(FF2) eq FldFin; 

    char := Characteristic(FF2);

    // Create unramified extension
  
    xl4 := CompleteFourThetaNullPointOverField(xl2);
    FF4 := BaseRing(Parent(xl4));

    Embed(FF2,FF4);

    RR := pAdicQuotientRing(char,prec);
    PZ := PolynomialRing(Integers());
    SS := UnramifiedExtension(RR,PZ!DefiningPolynomial(FF4));
    AssignNames(~SS,["t"]);
    KK := ResidueClassField(SS);

    Embed(FF4,KK);

    // Define modular equations

    P := PolynomialRing(Integers(),10);
    PP := PolynomialRing(Integers(),20);

    Riem := GetRiemannRelations(P);
    DRiem := Matrix([[ Derivative(f,i) : i in [2..10]] : f in Riem ]);
    ModEQs := GetCorrespondenceRelations(char,PP);
    DModEQs := Matrix([[ Derivative(f,i) : i in ([2..10] cat [12..20])] : f in ModEQs ]);    

    Xl4 := CoerceVector(xl4,SS);
    // Preprocess so that the Riemann equations are satisfied
    // to full precision:
    Xl2 := TwoTorsionPartOfFourThetaNullPoint(Xl4);
    Xl4 := CompleteFourThetaNullPointHensel(Xl2,Xl4);
    //
    Xl4 := NewtonThetaLift(Xl4,ModEQs,DModEQs,Riem,DRiem);
    return TwoTorsionPartOfFourThetaNullPoint(Xl4);
end intrinsic;


